/**
 * Copyright 1998-2007, CHISEL Group, University of Victoria, Victoria, BC, Canada.
 * All rights reserved.
 */
package ca.uvic.csr.shrimp.gui.QueryView;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.DefaultComboBoxModel;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JRootPane;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.ListSelectionModel;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.text.JTextComponent;

import ca.uvic.csr.shrimp.AttributeConstants;
import ca.uvic.csr.shrimp.FilterConstants;
import ca.uvic.csr.shrimp.DataBean.Artifact;
import ca.uvic.csr.shrimp.FilterBean.FilterBean;
import ca.uvic.csr.shrimp.ShrimpApplication.BeanNotFoundException;
import ca.uvic.csr.shrimp.ShrimpTool.ShrimpTool;
import ca.uvic.csr.shrimp.resource.ResourceHandler;
import ca.uvic.csr.shrimp.util.AutoCompleteDocument;
import ca.uvic.csr.shrimp.util.JButtonList;

/**
 * This class contains the components used in the query view.
 * 
 * @author Chris Callendar
 */
public class QueryPanelComponent extends JPanel implements ChangeListener {

	private static final String CONTAINS 	= "contains";
	private static final String STARTS_WITH = "starts with";
	private static final String ENDS_WITH	= "ends with";
	private static final String EXACT_MATCH = "exact match";
	private static final String REGEXP 		= "regexp";
	private static final String[] MODES = { CONTAINS, STARTS_WITH, ENDS_WITH, EXACT_MATCH, REGEXP };
	private static final int MAX_LEVEL = 5;
	private static final int MIN_LEVEL = -1;
	private static final int MAX_ROWS = 8;
	private static final int QUERY_ROW_HEIGHT = 24;
	
	private QueryView queryView;
	private QueryHelper queryHelper;
	private boolean updateGraphOnLevelsChange = false;

	private JRootPane rootPane;
	private JButtonList arcTypesList;
	private JButtonList nodeTypesList;
	private JButton btnQuery;
	private JLabel incomingLevelsLabel;
	private JLabel outgoingLevelsLabel;
	private JSpinner incomingLevels;
	private JSpinner outgoingLevels;
	private JButton showParentHierarchy;
	private JButton levelsHelp;
	private JComboBox searchBox;
	private JComboBox searchModes;
	private JPanel pnlSearchButtons;
	private AutoCompleteDocument autoCompleteDocument;
	private JRadioButton rdoUseSelected;
	private JRadioButton rdoSeachByName;
	
	private JPanel searchPanel;
	private JPanel nodesPanel;
	private JPanel arcsPanel;
	private JPanel levelsPanel;

	public QueryPanelComponent(QueryView queryView) {
		super();
		this.queryView = queryView;
		this.queryHelper = (queryView != null ? queryView.getQueryHelper() : new QueryHelper(null));	// for testing
		
		createGUI();
	}
	
	public void dispose() {
		this.queryView = null;
		this.queryHelper = null;
	}

	private void createGUI() {
		this.setLayout(new BorderLayout(0, 0));
		this.rootPane = new JRootPane();
		this.add(rootPane, BorderLayout.CENTER);
		rootPane.getContentPane().add(getSearchPanel(), BorderLayout.NORTH);

		JPanel mainPanel = new JPanel();
		mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.X_AXIS));
		mainPanel.add(getNodesPanel());
		mainPanel.add(Box.createHorizontalStrut(4));
		mainPanel.add(getArcsPanel());
		rootPane.getContentPane().add(mainPanel, BorderLayout.CENTER);
		rootPane.getContentPane().add(getLevelsPanel(), BorderLayout.EAST);

		updateValues();
		
		rootPane.setDefaultButton(btnQuery);
		getSearchBox().requestFocus();
		this.invalidate();
	}
	
	/**
	 * Adds or removes an attribute filter.
	 * @param attributeName the attribute name (see {@link AttributeConstants})
	 * @param filterType the filter type (see (see {@link FilterConstants})
	 * @param e the checkbox event
	 */
	private void filter(String attributeName, String filterType, ItemEvent e) {
		if (queryView != null) {
			try {
				FilterBean filterBean = (FilterBean) queryView.getBean(ShrimpTool.DISPLAY_FILTER_BEAN);
				boolean selected = (e.getStateChange() == ItemEvent.SELECTED);
				String item = e.getItem().toString();
				filterBean.addRemoveSingleNominalAttrValue(attributeName, String.class, filterType, item, !selected);
			} catch (BeanNotFoundException bnfe) {
				System.err.println("Couldn't find filter bean");
			}
		}
	}

	private void updateValues() {
		updateLevels();
		searchBox.setSelectedItem(queryHelper.getSrcArtifactName());
	}
	
	private int getStringMatchingMode() {
		String mode = (String) getSearchModesComboBox().getSelectedItem();
		if (CONTAINS.equals(mode)) {
			return QueryHelper.STRING_MATCH_CONTAINS_MODE;
		} else if (EXACT_MATCH.equals(mode)) {
			return QueryHelper.STRING_MATCH_EXACT_MODE;
		} else if (STARTS_WITH.equals(mode)) {
			return QueryHelper.STRING_MATCH_STARTS_WITH_MODE;
		} else if (ENDS_WITH.equals(mode)) {
			return QueryHelper.STRING_MATCH_ENDS_WITH_MODE;
		} else /* if (REGEXP.equals(mode)) */ {
			return QueryHelper.STRING_MATCH_REGEXP_MODE;
		}
	}
	
	private void doQuery() {
		if (queryView != null) {	
			// collect all information from gui widgets
			collectQueryParamsFromGUI();
			queryView.doQuery(true);
		}
	}

	private void collectQueryParamsFromGUI() {
		if (getUseSelectedRadioButton().isSelected()) {
			queryHelper.setSearchMode(QueryHelper.SEARCH_MODE_USE_SRC_ARTIFACTS);
		} else {
			queryHelper.setSearchMode(QueryHelper.SEARCH_MODE_BY_NAME);
			JTextComponent editorComponent = (JTextComponent) searchBox.getEditor().getEditorComponent();
			String text = editorComponent.getText().trim();
			queryHelper.setSrcArtifactName(text);
			queryHelper.setStringMatchingMode(getStringMatchingMode());
		}

		queryHelper.setArtifactTypes(nodeTypesList.getCheckedItemsCollection());
		queryHelper.setRelationshipTypes(arcTypesList.getCheckedItemsCollection());

		// levels are updated automatically when they change
	}

	private void setChecked(Collection types, JButtonList lst) {
		List indicesList = new ArrayList();
		lst.checkNone();
		for (int i = 0; i < lst.getModel().getSize(); i++) {
			String artType = (String) lst.getModel().getElementAt(i);
			if (types.contains(artType)) {
				indicesList.add(new Integer(i));
			}
		}
		int[] indices = new int[indicesList.size()];
		int i = 0;
		for (Iterator iter = indicesList.iterator(); iter.hasNext();) {
			Integer indexInt = (Integer) iter.next();
			indices[i] = indexInt.intValue();
			i++;
		}
		lst.setCheckedIndices(indices);
	}


	//////////////////////////////////////////////////////////////
	// PUBLIC METHODS
	//////////////////////////////////////////////////////////////
	
	public void stateChanged(ChangeEvent e) {
		// update the incoming and outgoing levels
		int incoming = ((Integer)incomingLevels.getValue()).intValue();
		int outgoing = ((Integer)outgoingLevels.getValue()).intValue();
		if ((incoming != queryHelper.getIncomingLevels()) || (outgoing != queryHelper.getOutgoingLevels())) {
			queryHelper.setLevels(incoming, outgoing);
			if (updateGraphOnLevelsChange) {
				doQuery();
			}
		}
	}
	
	public JPanel getSearchPanel() {
		if (searchPanel == null) {
			searchPanel = new JPanel(new BorderLayout());
			searchPanel.setBorder(BorderFactory.createTitledBorder("Search"));

			JPanel middlePanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 2, 0));
			middlePanel.add(getSearchByNameRadioButton());
			middlePanel.add(getSearchBox());
			pnlSearchButtons = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
			middlePanel.add(pnlSearchButtons);
			middlePanel.add(getSearchModesComboBox());
			middlePanel.add(getQueryButton());

			JPanel bottomPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 2, 0));
			bottomPanel.add(getUseSelectedRadioButton());

			searchPanel.add(middlePanel, BorderLayout.CENTER);
			searchPanel.add(bottomPanel, BorderLayout.SOUTH);
			
			ButtonGroup group = new ButtonGroup();
			group.add(getSearchByNameRadioButton());
			group.add(getUseSelectedRadioButton());
			
			getSearchByNameRadioButton().addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					boolean on = getSearchByNameRadioButton().isSelected();
					getSearchBox().setEnabled(on);
					searchModes.setEnabled(on);
					pnlSearchButtons.setEnabled(on);
				}
			});
		}
		return searchPanel;
	}

	public JComboBox getSearchModesComboBox() {
		if (searchModes == null) {
			searchModes = new JComboBox(MODES);
			searchModes.setPreferredSize(new Dimension(100, QUERY_ROW_HEIGHT));
		}
		return searchModes;
	}

	public JRadioButton getSearchByNameRadioButton() {
		if (rdoSeachByName == null) {
			rdoSeachByName = new JRadioButton("Search By Name ", true);
		}
		return rdoSeachByName;
	}

	public JRadioButton getUseSelectedRadioButton() {
		if (rdoUseSelected == null) {
			rdoUseSelected = new JRadioButton("Use Selected (none selected)", false);
			rdoUseSelected.setEnabled(false);
		}
		return rdoUseSelected;
	}

	public JComboBox getSearchBox() {
		if (searchBox == null) {
			searchBox = new JComboBox(new DefaultComboBoxModel());
			searchBox.setEditable(true);
	        autoCompleteDocument = new AutoCompleteDocument(searchBox);
	        JTextComponent editor = (JTextComponent) searchBox.getEditor().getEditorComponent();
			editor.setDocument(autoCompleteDocument);
	        
			searchBox.setPreferredSize(new Dimension(220, QUERY_ROW_HEIGHT));
			searchBox.addFocusListener(new FocusAdapter() {
				public void focusGained(FocusEvent e) {
					// pressing enter when textbox is focussed runs the query
					rootPane.setDefaultButton(btnQuery);
				}
			});
		}
		return searchBox;
	}

	public JButton getQueryButton() {
		if (btnQuery == null) {
			btnQuery = new JButton(new AbstractAction(" Query ", ResourceHandler.getIcon("icon_search.gif")) {
				public void actionPerformed(ActionEvent e) {
					doQuery();
				}
			});
			btnQuery.setToolTipText("Run the query");
			btnQuery.setPreferredSize(new Dimension(100, QUERY_ROW_HEIGHT));
		}
		return btnQuery;
	}
	
	public JPanel getNodesPanel() {
		if (nodesPanel == null) {
			nodesPanel = new JPanel(new BorderLayout());
			nodesPanel.setBorder(BorderFactory.createTitledBorder("Node Types"));
			nodeTypesList = new JButtonList();
			nodeTypesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			nodeTypesList.setListData(new Object[0]);
			nodesPanel.add(new JScrollPane(nodeTypesList), BorderLayout.CENTER);
			nodeTypesList.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					filter(AttributeConstants.NOM_ATTR_ARTIFACT_TYPE, FilterConstants.ARTIFACT_FILTER_TYPE, e);
				}
			});
		}
		return nodesPanel;
	}
	
	public JPanel getArcsPanel() {
		if (arcsPanel == null) {
			arcsPanel = new JPanel(new BorderLayout());
			arcsPanel.setBorder(BorderFactory.createTitledBorder("Arc Types"));
			arcTypesList = new JButtonList();
			arcTypesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			arcTypesList.setListData(new Object[0]);
			arcsPanel.add(new JScrollPane(arcTypesList), BorderLayout.CENTER);
			arcTypesList.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					filter(AttributeConstants.NOM_ATTR_REL_TYPE, FilterConstants.RELATIONSHIP_FILTER_TYPE, e);
				}
			});
		}
		return arcsPanel;
	}
	
	public JPanel getLevelsPanel() {
		if (levelsPanel == null) {
			levelsPanel = new JPanel(new BorderLayout());
			levelsPanel.setBorder(BorderFactory.createTitledBorder("Levels"));

			JPanel topPanel = new JPanel(new GridLayout(2, 1, 0, 1));
			levelsPanel.add(topPanel, BorderLayout.NORTH);
			
			JPanel pnl = new JPanel(new FlowLayout(FlowLayout.LEFT, 2, 2));
			pnl.add(getIncomingLevelsLabel());
			pnl.add(getIncomingLevelsSpinner());
			pnl.add(Box.createHorizontalStrut(4));
			pnl.add(getShowHierarchyButton());
			topPanel.add(pnl);
			
			pnl = new JPanel(new FlowLayout(FlowLayout.LEFT, 2, 2));
			pnl.add(getOutgoingLevelsLabel());
			pnl.add(getOutgoingLevelsSpinner());
			pnl.add(Box.createHorizontalStrut(4));
			pnl.add(getLevelsHelpButton());
			topPanel.add(pnl);
		}
		return levelsPanel;
	}

	/**
	 * @return the incoming levels spinner - allowed values are from 
	 * 	{@link QueryPanelComponent#MIN_LEVEL} to {@link QueryPanelComponent#MAX_LEVEL}.
	 */
	public JSpinner getIncomingLevelsSpinner() {
		if (incomingLevels == null) {
			incomingLevels = new JSpinner(new SpinnerNumberModel(1, MIN_LEVEL, MAX_LEVEL, 1));
			incomingLevels.addChangeListener(this);
		}
		return incomingLevels;
	}

	/**
	 * @return the incoming levels spinner - allowed values are from 
	 * 	{@link QueryPanelComponent#MIN_LEVEL} to {@link QueryPanelComponent#MAX_LEVEL}.
	 */
	public JSpinner getOutgoingLevelsSpinner() {
		if (outgoingLevels == null) {
			outgoingLevels = new JSpinner(new SpinnerNumberModel(1, MIN_LEVEL, MAX_LEVEL, 1));
			outgoingLevels.addChangeListener(this);
		}
		return outgoingLevels;
	}
	
	public JLabel getIncomingLevelsLabel() {
		if (incomingLevelsLabel == null) {
			incomingLevelsLabel = new JLabel("Incoming ");
			incomingLevelsLabel.setPreferredSize(new Dimension(80, 16));
		}
		return incomingLevelsLabel;
	}

	public JLabel getOutgoingLevelsLabel() {
		if (outgoingLevelsLabel == null) {
			outgoingLevelsLabel = new JLabel("Outgoing ");
			outgoingLevelsLabel.setPreferredSize(new Dimension(80, 16));
		}
		return outgoingLevelsLabel;
	}	
	
    /**
     * Returns the search by name panel.
     * You can add buttons to this panel - it is located to the right 
     * of the search box.
     */
    public JPanel getSearchByNamePanel() {
    	return pnlSearchButtons;
    }
    
	/**
	 * Returns the show hierarchy button (next to the incoming levels spinner).
	 */
	public JButton getShowHierarchyButton() {
		if (showParentHierarchy == null) {
			showParentHierarchy = new JButton(new AbstractAction(null, ResourceHandler.getIcon("icon_parent_hierarchy.gif")) {
				public void actionPerformed(ActionEvent e) {
					incomingLevels.setValue(new Integer(MIN_LEVEL));
				}
			});
			showParentHierarchy.setPreferredSize(new Dimension(22, 20));
			showParentHierarchy.setToolTipText("Show the parent hierarchy up to the root node");
		} 
		return showParentHierarchy;
	}

	/**
	 * Returns the show hierarchy button (next to the incoming levels spinner).
	 */
	public JButton getLevelsHelpButton() {
		if (levelsHelp == null) {
			final Icon icon = ResourceHandler.getIcon("icon_help.gif");
			levelsHelp = new JButton(new AbstractAction(null, icon) {
				public void actionPerformed(ActionEvent e) {
					StringBuffer msg = new StringBuffer();
					msg.append("For incoming and outgoing choose an integer between " + MIN_LEVEL + " and " + MAX_LEVEL + ".\n");
					msg.append("This controls how many levels to display away from the matched nodes.\n");
					msg.append("The higher the number the more nodes will be displayed.\n\n");
					msg.append("For immediate neighbours only set incoming and outgoing to 1.\n");
					msg.append("For the parent hierarchy set incoming to -1 and outgoing to 0.\n");
					JOptionPane.showMessageDialog(levelsHelp, msg.toString(), "Levels Help", JOptionPane.QUESTION_MESSAGE);
				}
			});
			levelsHelp.setPreferredSize(new Dimension(22, 20));
			levelsHelp.setToolTipText("Help");
		} 
		return levelsHelp;
	}
	
	public void setSearchItems(Collection items, boolean clearExisting) {
		DefaultComboBoxModel model = (DefaultComboBoxModel) getSearchBox().getModel();
		if (clearExisting) {
			model.removeAllElements();
		}
		for (Iterator iter = items.iterator(); iter.hasNext(); ) {
			model.addElement(iter.next());
		}
	}
	
	public void addSearchItem(Object item) {
		DefaultComboBoxModel model = (DefaultComboBoxModel) getSearchBox().getModel();
		model.addElement(item);
	}

	public void addSearchItem(int index, Object item) {
		DefaultComboBoxModel model = (DefaultComboBoxModel) getSearchBox().getModel();
		if ((index >= 0) && (index < model.getSize())) {
			model.insertElementAt(item, index);
		}
	}
	
	public void updateSearchMode() {
		int mode = queryHelper.getSearchMode();
		String strMode = (String) getSearchModesComboBox().getSelectedItem();
		if (QueryHelper.STRING_MATCH_CONTAINS_MODE == mode) {
			strMode = CONTAINS;
		} else if (QueryHelper.STRING_MATCH_EXACT_MODE == mode) {
			strMode = EXACT_MATCH;
		} else if (QueryHelper.STRING_MATCH_STARTS_WITH_MODE == mode) {
			strMode = STARTS_WITH;
		} else if (QueryHelper.STRING_MATCH_ENDS_WITH_MODE == mode) {
			strMode = ENDS_WITH;
		} else if (QueryHelper.STRING_MATCH_REGEXP_MODE == mode) {
			strMode = REGEXP;
		}
		getSearchModesComboBox().setSelectedItem(strMode);
	}
	
	public void updateLevels() {
		incomingLevels.setValue(new Integer(queryHelper.getIncomingLevels()));
        outgoingLevels.setValue(new Integer(queryHelper.getOutgoingLevels()));
	}

	public void updateArtifactTypes() {
		nodeTypesList.setListData(new Vector(queryHelper.getArtTypes()));
        setChecked(queryHelper.getArtTypes(), nodeTypesList);
	}

	public void updateRelationshipTypes() {
		arcTypesList.setListData(new Vector(queryHelper.getRelTypes()));
        setChecked(queryHelper.getRelTypes(), arcTypesList);
	}
	
	public void updateSrcArtifacts() {
		Collection srcArtifacts = queryHelper.getSrcArtifacts();
		boolean hasSrcArtifacts = !srcArtifacts.isEmpty();
		if (hasSrcArtifacts) {
			StringBuffer names = new StringBuffer();
			for (Iterator iter = srcArtifacts.iterator(); iter.hasNext(); ) {
				Artifact artifact = (Artifact) iter.next();
				if (names.length() > 0) {
					names.append(", ");
				}
				names.append(artifact.getName());				
			}
			getUseSelectedRadioButton().setSelected(true);
			getUseSelectedRadioButton().setEnabled(true);
			getUseSelectedRadioButton().setText("Use Selected (" + names.toString() + ")");
        } else {
        	getSearchByNameRadioButton().setSelected(true);
        	getUseSelectedRadioButton().setEnabled(false);
			getUseSelectedRadioButton().setText("Use Selected (none selected)");
        }
	}
	
	public void updateSrcArtifactName() {
		if (queryHelper.getSearchMode() == QueryHelper.SEARCH_MODE_BY_NAME) {
			getSearchByNameRadioButton().setSelected(true);
		}
		searchBox.setSelectedItem(queryHelper.getSrcArtifactName());
		JTextComponent textbox = (JTextComponent)searchBox.getEditor().getEditorComponent();
		// this is to stop the UI from freezing when the popup shows
		autoCompleteDocument.setShowPopup(false);
		textbox.setText(queryHelper.getSrcArtifactName());
		autoCompleteDocument.setShowPopup(true);
	}

	public void focusSearchBox() {
		JTextComponent editorComponent = (JTextComponent) searchBox.getEditor().getEditorComponent();
		editorComponent.requestFocus();
		editorComponent.selectAll();
	}
	
	public void hideSearchRadioButtons() {
		getUseSelectedRadioButton().setVisible(false);
		getSearchByNameRadioButton().setVisible(false);
		getSearchPanel().revalidate();
	}

	public void showSearchRadioButtons() {
		getUseSelectedRadioButton().setVisible(true);
		getSearchByNameRadioButton().setVisible(true);
		getSearchPanel().revalidate();
	}
	
	/**
	 * If update is true then whenever the incoming or outgoing levels change value
	 * the graph is updated to display the new levels.
	 * @param update
	 */
	public void setUpdateGraphWhenLevelsChange(boolean update) {
		this.updateGraphOnLevelsChange = update;
	}

	/**
	 * Tries to make the height better depending on what components are shown
	 * and how many items are in the lists.
	 */
	public void adjustHeight() {
		int height = 0;
		if (getSearchPanel().isVisible()) {
			height += (getUseSelectedRadioButton().isVisible() ? 75 : 60);
		}
		int middleHeight = 80;
		int rows = 0;
		if (getNodesPanel().isVisible()) {
			rows = Math.min(MAX_ROWS, nodeTypesList.getModel().getSize());
		}
		if (getArcsPanel().isVisible()) {
			rows = Math.min(MAX_ROWS, Math.max(rows, arcTypesList.getModel().getSize()));
		}
		middleHeight = Math.max(middleHeight, 16 + (rows * 18));
		height += middleHeight;
		this.setPreferredSize(new Dimension(this.getPreferredSize().width, height));
	}
	
	public static void main(String[] args) {
		QueryPanelComponent qpc = new QueryPanelComponent(null);
		
		JFrame frame = new JFrame("QueryPanelComponent");
		frame.setSize(750, 500);
		frame.setPreferredSize(new Dimension(750, 500));
		frame.setLocation(400, 200);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.getContentPane().add(qpc, BorderLayout.NORTH);
		frame.pack();
		frame.setVisible(true);		
	}

}